################################################################################
# Created       : Martin Foster
# Modified      : 22-Feb-2006
################################################################################
#
# Comm - Module part of Ethereal Realms, designed to prompt and handle posts.
# Copyright (C) 2000-2006  Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 29-26503 TWP RD 511
#                 Spruce Grove, Alberta
#                 T7Y 1G4
#
#################################################################################

package Ethereal::Comm;						# Pacakace name

use Ethereal::Filter;						# Filter system
use Ethereal::Mail;						# Mail sending module
use Ethereal::Option;						# Options handling
use Ethereal::Stamp;						# Timestamp handling

use strict;


################################################################################
# Sub-Routines
################################################################################

#####################
# Constructor
#
# Allows to class to be exported and thus spawn objects as a result

sub new
{
	#####################
	# Data Members
	my $class    = shift;					# Self reference
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI Handle
	my $tmpl     = shift;					# Templates

	my $self     = {
		HANDLE   => $database,
		CGI      => $cgi,
		TEMPLATE => $tmpl
 	 };


	#####################
	# Program Area

	# Bless $self into designated class
	bless $self, $class;

	# Return self
	return $self;
}



#####################
# Comm
#
# Subroutine primarily charged with displaying input area.   This will also call
# CommHandler which will handle the requests to determin the need for a new
# database entry

sub Comm
{
	#####################
	# Data members
	my $self     = shift;					# Self referencing link
	my $login    = shift;					# Auth/Header handler
	my $param    = shift;					# Parameter handle
	my $choice   = shift;					# Choice of the matter

	my $database = $self->{'HANDLE'};			# Database handle
	my $cgi      = $self->{'CGI'};				# CGI handle

	my $js;							# Javascript header
	my $jscode;						# Javascript code
	my $jslink   = '';					# Javascript link code
	my $jsstrm;						# Javascript stream code
 
	my $header;						# Independant header
	my $panel;						# Communications panel

	my $basic;						# Basic parameter list
	my $inline;						# Inline parameter list
	my $url      = $cgi->url(-absolute=>1);			# URL of script

	my $comm     = $param->{'COMM'};			# Comm value
	my $color;						# Colour default
	my $default;						# Default colour
	my $locate;						# Location
	my $face;						# Font face
	my $puppet;						# Puppet name
	my $size;						# Size default
	my $text;						# Text colour

	my $access;						# Access control
	my $locale;						# Default settings
	my $snub;						# Limiting groups

	my @color;						# Colours
	my @face;						# Font face
	my @locale;						# Locale settings
	my @size;						# Size fonts
	my @char;						# Character

	my %realm;						# Realm hash


	#####################
	# Program area

	# Link with Hash
	$database->GetHashRealm(\%realm);

	# Locale setting
	($access) = $database->DataGetRealmAccess($param->{'ROOM'});
	($locale) = $database->DataGetRealmLocale($param->{'ROOM'});


	# User settings
	($js)   = $database->DataGetJavascript($param->{'USER'});
	($snub) = $database->DataGetSnub($param->{'USER'});


	# Store character name
	$puppet = $param->{'CHAR'};

	# Populate arrays
	@color  = FontSelectColor($self, $param);
	@size   = FontSelectSize($self, $param);
	@face   = FontSelectFace($self, $param);

	# Sorting
	@color  = sort(@color);


	# Pupulate puppet list
	$database->GroupList(\@char, $access, $param->{'USER'}, $param->{'ROOM'}, $snub, $puppet);

	# Parameter cleanup
	$param->Cleanup(
	  'COLOR',
	  'COMM',
	  'FACE',
	  'POST',
	  'SIZE',
	  'SNUB'
	 );


	# Inline parameter handling
	# Basic inline
	$basic  = $param->EmbedInline(
		USER  => $param->{'USER'},
		CRYPT => $param->{'CRYPT'}
	 );

	# Normal use inline
	$inline = $param->EmbedInline($param->Flat());

	# Pull realm default
	($default) = $database->DataGetRealmText($param->{'ROOM'});	

	# Retreive size and color
	($size, $color, $face) = $database->DataGetFont($puppet, $param->{'USER'}, $default, $face[0]);


	# Independent HTML
	if ($choice != 0)
	{
		# Extract header
		$header = $database->FetchRealmHeader($param->{'ROOM'});

		# Focus handling
		# Extraction and replacement
		if (($js eq 'yes')
		 && ($header =~ /<(BODY.*?)>/si))
		{
			# Safety mechanism
			# Assign
			my $org = $1;
			my $new = $1;

			# Prepare
			$new = quotemeta($new);

			# Replace
			$header =~ s/$new/$org ONLOAD="document.SEND.POST.focus();"/gs;
		}
			

		# Print file and HTML header
		$login->WriteHeader();
		print $header, "\n";
	}

	# Javascript handling
	# Print if javascript is desired
	if ($js eq 'yes')
	{ 
		# Assign values
		$jscode = $database->FetchRealmJavascript($param->{'ROOM'}); 

		# Generate javascript code
		$jslink = "ONCLICK='window.open(this.href,this.target,settings); return false;'";


		# Internet explorer and excluding basic
		if (($cgi->user_agent('MSIE'))
		 && ($choice != 0))
		{
			# Set variables
			my $url  = $cgi->url(-absolute=>1);
			my $time = substr(time, -5, 4);

			# Pull widget
			$jsstrm = $database->FetchRealmSupp($param->{'ROOM'});


			# Generate links
			if ((defined($comm))
			 && ($comm =~ /false/i))
			{
				# Attach link
				$jsstrm =~ s/LATTACH/HREF=\"$url$inline&COMM=True\" ACCESSKEY=\"a\" ONCLICK=\"AttachFrames(this.href); return false;\"/gs;
			}
			else
			{
				# Target time 
				my $time = time;

				# Detach link
				$jsstrm =~ s/LATTACH/HREF=\"$url$inline&COMM=False\" TARGET=\"$time\" ACCESSKEY=\"d\" ONCLICK=\"DetachFrames(this.href,this.target); return false;\"/gs;
			}
		}
		# Display code
		print "$jscode\n";
	}

	# Retreive appropriate panel
	($panel) = (defined($param->{'OFFLINE'}))
	  ? $database->FetchOfflinePanel($param->{'ROOM'})
	  : $database->FetchRealmPanel($param->{'ROOM'});


	# General link handling
	# Username/password only
	$panel =~ s/BASEPARAM/$basic/gs;

	# Code necessary for popups
	$panel =~ s/GENJSCODE/$jslink/gs;

	# All necessary realm continuance
	$panel =~ s/GENPARAM/$inline/gs;
	

	# Determine colour
	# Initial undefined assignment
	my $wfontcolr;

	# Loop
	foreach $text (@color)	
	{
		# Match made
		if ($text eq $color)
		{
			# Data assignment
			$wfontcolr = $cgi->popup_menu(
				-name      => 'COLOR',
				-values    => \@color,
				-default   => $color,
				-tabindex  => 20,
				-class     => 'WFONTCOLR');

			# Exit loop
			last;
		}
	}

	# Assign default otherwise
	$wfontcolr = $cgi->popup_menu(
		-name      => 'COLOR',
		-values    => \@color,
		-default   => $default,
		-tabindex  => 20,
		-class     => 'WFONTCOLR'
	 ) unless (defined($wfontcolr));


	# Widgets
	# Assigning values
	my $wpostsend = $cgi->submit(
		-name      => $realm{'TxtPost'},
		-value     => $realm{'TxtPost'},
		-tabindex  => 3,
		-class     => 'WPOSTSEND',
		-accesskey => 'p');

	# Posting window
	my $wposttext = $cgi->textarea(
		-name      => 'POST',
		-rows      => 3,
		-columns   => 50,
		-wrap      => 'SOFT',
		-tabindex  => 2,
		-class     => 'WPOSTTEXT',
		-accesskey => 't');

	# Popup menus
	# Character select
	my $wcharslct = $cgi->popup_menu(
		-name      => 'CHAR',
		-values    => \@char,
		-default   => $puppet,
		-class     => 'WCHARSLCT',
		-tabindex  => 1);

	# Font face selct
	my $wfontface = $cgi->popup_menu(
		-name      => 'FACE',
		-values    => \@face,
		-default   => $face,
		-tabindex  => 21,
		-class     => 'WFONTFACE');

	# Font size select
	my $wfontsize = $cgi->popup_menu(
		-name      => 'SIZE',
		-values    => \@size,
		-default   => $size,
		-tabindex  => 22,
		-class     => 'WFONTSIZE');

	# Active realms
	# Snubbing
	my $wsnubctrl = ($snub eq 'yes')
	  ? $cgi->checkbox(-name=>'SNUB', -label=>'', -checked=>'on')
	  : $cgi->checkbox(-name=>'SNUB', -label=>'');
	

	# Scroll back functionality
	my $wscrladvd = ($choice < 1)
	  ? $cgi->textfield('SCROLLBACK', '', 3, 3) . " " . $realm{'TxtScrollback'}
	  : '&nbsp;';
	my $wscrlback = ($choice < 1)
	  ? $cgi->checkbox(-name=>'SCROLLBACK', -label=>$realm{'TxtScrollback'})
	  : '&nbsp;';
	my $wscrlctrl = ($choice < 1)
	  ? $cgi->checkbox(-name=>'GETPOSTS',   -label=>$realm{'TxtRemote'})
	  : '&nbsp;';
	my $wstrmctrl = ($choice > 1)
	  ? $cgi->checkbox(-name=>'STRMKILL',   -label=>$realm{'TxtStream'})
	  : '&nbsp;';


	# Locale settings
	if ($locale ne 'none')
	{
		# Bring into play
		my $wcharlocl;
		my $wcharfilt;


		# Filtering widget
		$wcharfilt = (defined($param->{'ICPOSTS'})) 
			? $cgi->checkbox(-name=>'ICPOSTS', -label=>'', -checked=>'on')
			: $cgi->checkbox(-name=>'ICPOSTS', -label=>'');

		# Appropriate cleanup
		$param->Cleanup('ICPOSTS');



		# Locale widget
		# Current location
		($locate) = $database->DataGetPuppetLocale($puppet, $param->{'USER'});


		# Strict list
		if ($locale eq 'strict')
		{
			# Location defined
			$locate = (defined($locate)) ? $locate : $realm{'TxtNone'};


			# Pull list of locales
			$database->GetListRealmLocale(\@locale, $param->{'ROOM'});

			# Append default
			unshift(@locale, $realm{'TxtNone'});

			# Generate widget
			$wcharlocl = $cgi->popup_menu('LOCL', \@locale, $locate);
		}

		# Simple textfield
		else
		{
			# Location defined
			$locate = (defined($locate)) ? $locate : '';

			# Generate widget
			$wcharlocl = $cgi->textfield('LOCL', $locate, 25, 50);
		}

		# Replace macro
		$panel =~ s/WCHARLOCL/$wcharlocl/gs;
		$panel =~ s/WCHARFILT/$wcharfilt/gs;
	}


	# Widget replacements
	# Posting widgets
	$panel =~ s/WPOSTSEND/$wpostsend/gs;
	$panel =~ s/WPOSTTEXT/$wposttext/gs;

	# Puppet selection
	$panel =~ s/WCHARSLCT/$wcharslct/gs;

	# Font selection
	$panel =~ s/WFONTCOLR/$wfontcolr/gs;
	$panel =~ s/WFONTFACE/$wfontface/gs;
	$panel =~ s/WFONTSIZE/$wfontsize/gs;

	# Specific controls
	# Regular active realms
	unless (defined($param->{'OFFLINE'}))
	{
		# Snubbing
		$panel =~ s/WSNUBCTRL/$wsnubctrl/gs;
	
		# Scrollback/Archiving/Stream Kill
		$panel =~ s/WSCRLADVD/$wscrladvd/gs;
		$panel =~ s/WSCRLBACK/$wscrlback/gs;
		$panel =~ s/WSCRLCTRL/$wscrlctrl/gs;
		$panel =~ s/WSTRMCTRL/$wstrmctrl/gs;
	}


	# Framed, streaming, autoscrolling 
	if ($choice != 0)
	{
		# Frame sensitive information
		$param->{'COMM'} = (defined($comm)) ? $comm : 'True';
	}


	# HTML
	# Start form
	print $cgi->start_form(-name=>'SEND');
	print $param->EmbedNormal($param->Flat()), "\n"; 

	# Display panel
	print "$panel\n";

	# End of form
	print $cgi->end_form(), "\n";


	# Print end of html if needed
	if ($choice != 0)
	{ 
		# Internet explorer
		if ($cgi->user_agent('MSIE'))
		{
			# Display detached stream
			print $jsstrm, "\n";
		}


		# Retreive proper footer
		print $database->FetchRealmFooter($param->{'ROOM'}); 
	}
}

# Alias for dependency
sub CommDependant
{ Comm(shift, shift, shift, 0); }

# Alias for independnt framed
sub CommIndependant
{ Comm(shift, shift, shift, 1); }

# Alias for independant streaming
sub CommStream
{ Comm(shift, shift, shift, 2); }

# Alias for independant auto-scrolling streaming
sub CommAutoScroll
{ Comm(shift, shift, shift, 3); }


#####################
# DataWrite
#
# Encharged with properly handling information and actually making posts
# once formatted inserting them into the database.

sub DataWrite
{
        #####################
        # Data members
	my $self     = shift;					# Self referencing link
	my $param    = shift;					# Parameter handle
	my $chatpref = shift;					# Chat preferences

	my $database = $self->{'HANDLE'};			# Database handle

	my $class    = 'special';				# Class updates
	my $cmd      = 'stream';				# Command
	
	my $locate;						# Location handler
	my $puppet;						# Puppet name and existence
	my $snub;						# Snubbing

	my %realm;						# Realm hash


        #####################
        # Data members

	# Link with hash
	$database->GetHashRealm(\%realm);


	# Make sure its defined
	$chatpref = (defined($chatpref)) ? $chatpref : 'basic';

	
	# Assign to target if undefined
	$param->{'TARGET'} = (defined($param->{'TARGET'}))
	  ? $param->{'TARGET'}
	  : 'all';

	# Make sure puppet is valid
	($puppet) = $database->DataGetPuppetExistence($param->{'CHAR'}, $param->{'USER'});


	# Safe stream enhancements
	# Make sure
	if ((!defined($param->{'PSTART'}))
	 && (!defined($param->{'OFFLINE'})))
	{
		# Always pull the next sequential number and work from there
		($param->{'PSTART'}) = $database->DataGet("SELECT NEXTVAL('seqPost')");

		# Only apply if necessary
		if (($chatpref == 'stream')
		 || ($chatpref == 'autoscroll'))
	 	{
			# Kill off previous breeds of streams
			$database->Quick("INSERT INTO Post 
				(PostIDNumber,
				 RealmName, 
				 PuppetName,
				 PuppeteerLogin,
				 PostTo,
				 PostClass,
				 PostCmd,
				 PostFullFormat,
				 PostImagelessFormat,
				 PostPartialFormat)
				VALUES(?,?,?,?,?,?,?,?,?,?)", 

			 $param->{'PSTART'},
			 $param->{'ROOM'}, 
			 $param->{'CHAR'},
			 $param->{'USER'},
			 $param->{'USER'},
	 		 $class,
	 		 $cmd,
	 		 $realm{'TmplInitial'},
	 		 $realm{'TmplInitial'},
			 $realm{'TmplInitial'});	
		}
 	}
		

	# Puppet must exist
	if (defined($puppet))
	{
		# Only do following if posts exist
		if ((defined($param->{'POST'}))
		 && ($param->{'POST'} =~ /.{3,}/))
		{
			# Update if not a dice roll or system wide post
			if (($param->{'TARGET'} ne 'dice roll')
			 && ($param->{'TARGET'} ne $realm{'TxtSystem'}))
			{
				# Determine location
				# Location handling
				if (defined($param->{'LOCL'}))
				{
					# Valid entry
					if (($param->{'LOCL'} ne $realm{'TxtNone'})
					 && (length($param->{'LOCL'}) > 2))	
					{
						# Direct association
						($locate) = $param->{'LOCL'};	
					}

					# Invalid
					else
					{
						# Cleanup 
						$param->Cleanup('LOCL');
					}
				}

				# Re-assign defaults
				else
				{
					# Use defaults unless you can reset
					($locate) = $database->DataGetPuppetLocale($param->{'CHAR'}, $param->{'USER'});

					# Cleanup
					$param->Cleanup('LOCL');
				}

				# Sanity check
				$locate = substr(DBI::neat($locate, 52), 1, -1) if (defined($locate));
				

				# Update database entry
				$database->Quick("UPDATE Puppet 
					SET PuppetTextColour=?, 
					    PuppetTextSize=?,
					    PuppetTextFace=?, 
					    PuppetLocale=?
					WHERE PuppetName=?
					  AND PuppeteerLogin=?", 

				 $param->{'COLOR'}, 
			 	 $param->{'SIZE'},
				 $param->{'FACE'}, 
				 $locate,
				 $param->{'CHAR'},
				 $param->{'USER'});


			 	# Confirm snubbing
				# Safety
				$snub = (defined($param->{'SNUB'}))
				  ? 'yes'
				  : 'no';
				
				# Update snubbing
				$database->Quick("UPDATE Puppeteer
					SET PuppeteerSnub=?
					WHERE PuppeteerLogin=?",

				 $snub,
				 $param->{'USER'}
				);
 			 }

			# Handle posts
			$self->PostHandler($param, \%realm); 
		}
	
		# No post simply post for timestamp purposes
		# Active realms only
		elsif (!defined($param->{'OFFLINE'}))
		{
			# Make people aware their presence
			$database->Quick("INSERT INTO Post 
				(RealmName, 
				 PuppetName,
				 PuppeteerLogin,
				 PostTo)
				VALUES(?,?,?,'system')", 

			 $param->{'ROOM'}, 
			 $param->{'CHAR'},
			 $param->{'USER'});
		}
	}
}


#####################
# FontSelect
#
# Based on the ability to allow hosts to select which colours, fonts faces and 
# sizes they wish to allow.  This sub-routine will determin what to send
# back.  As being either the hosts entity or a complete list as default.

sub FontSelect
{
        #####################
        # Data members
        my $self     = shift;                                   # Self referencing link
        my $param    = shift;                                   # Parameter handle
	my $type     = shift;					# Font type

        my $database = $self->{'HANDLE'};                       # Database handle

	my $options;						# String of options
	my @options;						# List of options


	#####################
	# Program Area

	# Pull realm values	
	($options) = $database->DataGet("SELECT RealmFont$type 
		FROM  Realm
		WHERE RealmName=?", $param->{'ROOM'});

	# Valid or not
	unless (defined($options))
	{
		# Retreive list
		$database->GetList(\@options, "SELECT FontName
			FROM Font
			WHERE FontType=?
			ORDER BY FontCode", lc($type));
	}
	else
	{
		# Seperate into useable text
		@options = split(/:/, $options);
	}

	# Return defaults
	return @options;
}

# Alias for font color
sub FontSelectColor
{ return FontSelect(shift, shift, 'Color'); }

# Alias for font face
sub FontSelectFace
{ return FontSelect(shift, shift, 'Face'); }

# Alias for font size
sub FontSelectSize
{ return FontSelect(shift, shift, 'Size'); }


####################
# Gummy Handler
#
# This will control how a gummy is shown.   This will vary depending on access
# and which real they access at the time.  Since such behavior is for 
# moderators or supervisors only this will not be triggered often.

sub GummyHandler
{
	#####################
	# Data members
	my $self     = shift;					# Self referencing link
	my $database = shift;					# Database handle
	my $param    = shift;					# Parameter handle
	my $realm    = shift;					# Realm handle
	my $pup      = shift;					# Results page

	my $acl;						# Choice of the matter
	my $class;						# Gummy class
	my $gummy;						# Gummy returned
	my $super    = 16;					# Supervisory code

	#####################
	# Program Area

	# Only if defined
	if (defined($pup->{'PuppetGummy'}))
	{
		# Retreive gummy class
		($class) = $database->DataGetGummyClass($pup->{'PuppetGummy'});

		# Need to continue with check
		if (($class eq 'super')
		 or ($class eq 'mod'))
		{
			# Additional checks based on ACL
			# Retrieve access control lists
			($acl) = $database->DataGetPuppeteerControl($param->{'ROOM'}, $param->{'USER'});

			# User fails a check
			unless ((defined($acl))
			   and ((($class eq 'super') && ($acl >= $super))
			    or  (($class eq 'mod')   && ($acl < $super))))
			{
				# Reset to default gummy
				$gummy = $realm->{'SetGummyDefault'};
			}
		}
	}
		
	# Gummy assignment
	$gummy = (defined($gummy)) 
	 ? $gummy 
	 : $realm->{'SetGummyDefault'};  

	# Generate bio
	$gummy = ($gummy ne 'NIL')
	 ? "MLINK<IMG SRC=\"$gummy\" BORDER=\"0\" ALIGN=\"LEFT\" ALT=\"$param->{CHAR}\"></A>"
	 : '';
	
	# Return
	return $gummy;
}


#####################
# Post Handler
#
# Depending on rules will handle the formatting of the three styles of posts
# filter language and HTML if needed or specified.

sub PostHandler
{
	#####################
	# Data members
	my $self     = shift;					# Self referencing link
	my $param    = shift;					# Parameter handle
	my $realm    = shift;					# Reference to realm

	my $cgi      = $self->{'CGI'};				# CGI handle
	my $database = $self->{'HANDLE'};			# Database handle
	my $tmpl     = $self->{'TEMPLATE'};			# Template handling

	my $p_handle;						# Puppet information
	my $r_handle;						# Realm information
	my $t_handle;						# Target information

	my $filter   = new Ethereal::Filter();			# Filter handler
	my $option   = new Ethereal::Option();			# Option handler

	my $res;						# General results
	my $pup;						# Puppet results
	my $rlm;						# Realm results

	my $full;						# Full formating
	my $imageless;						# No images
	my $spartan;						# Spartan formatting

	my $img_avg;						# Average shrinking
	my $img_dis;						# Imake itself
	my $img_height;						# Image height
	my $img_lnk;						# Image link
	my $img_width;						# Image width
	my $img_size;						# Image size

	my $cmd;						# Embeded post commands

	my $class;						# Class handler
	my $dest     = 'all';					# Target
	my $gummy;						# Gummy link
	my $inline;						# Inline parameters
	my $name;						# Formatted name
	my $notice;						# Notice Emails
	my $window;						# Window name to open

	my $tag;						# Tagline
	my $title;						# Title

	my $adm;						# Administrative
	my $mod;						# Moderator
	my $sup;						# Supervisory

	my $count    = 0;					# Simple counter
	my $offline;						# Offline post extract
	my $post;						# Actual post
	my $position = 0;					# Position of post

	my $target   = $param->{'TARGET'};			# Target to

	my @lang;						# Language array
	my @posts;						# Array of posts

	my %search;						# Search and replace hash
	my %f_search;						# Full viewing
	my %i_search;						# Imageless mode
	my %s_search;						# Spartan mode


	#####################
	# Program area

	# Initial configuration
	%search = (
		MPCLASS   => 'PSTPRI',				# Default class
		MCMTHEAD  => '<!-HEAD-->',			# Delimit header
		MCMTPOST  => '<!-POST-->',			# Delimit post
		MNOTICE   => ''					# Additional notice
	);
	
	
	# Post commands
	$cmd = (defined($param->{'ICPOSTS'}))  ? 'play'   : 'general';
	$cmd = (defined($param->{'STRMKILL'})) ? 'stream' : $cmd;


	# Retreive puppet and realm information
	# Puppet information
	$database->Pull(\$p_handle, "SELECT 
		  PuppetName      AS \"PuppetName\",
		  PuppetPool	  AS \"PuppetPool\",
		  PuppeteerLogin  AS \"PuppeteerLogin\", 
		  PuppetTag       AS \"PuppetTag\",
		  PuppetPic       AS \"PuppetPic\",
		  PuppetPicHeight AS \"PuppetPicHeight\",
		  PuppetPicWidth  AS \"PuppetPicWidth\",
		  PuppetPicSize   AS \"PuppetPicSize\",
		  PuppetTitle     AS \"PuppetTitle\", 
		  PuppetGummy     AS \"PuppetGummy\"
		FROM  Puppet 
		WHERE PuppetName=?
		  AND PuppeteerLogin=?",

	 $param->{'CHAR'},
	 $param->{'USER'}
	);

	# Realm Information
	$database->Pull(\$r_handle, "SELECT 
		  RealmAccess      AS \"RealmAccess\",
		  RealmLanguage    AS \"RealmLanguage\", 
		  RealmFilter      AS \"RealmFilter\",
		  RealmTags        AS \"RealmTags\",
		  RealmDateLoc     AS \"RealmDateLoc\",
		  RealmImageHeight AS \"RealmImageHeight\",
	          RealmImageWidth  AS \"RealmImageWidth\",
	          RealmImageSize   AS \"RealmImageSize\"
		FROM  Realm 
		WHERE RealmName=?",

	 $param->{'ROOM'}
	);

	# Associate to hash
	$pup = $p_handle->fetchrow_hashref();
	$rlm = $r_handle->fetchrow_hashref(); 	


	# Frame border and class
	# Offline realms
	if (defined($param->{'OFFLINE'}))
	{
		# Set search
		$search{'MPCLASS'} = 'PSTOLN';
		
		# Active realm prep
		$class = 'offline';

		# Quickly copy and strip post
		$offline = $filter->StripHTML($param->{'POST'});
	}
	
	# Standard realms
	else
	{
		# General posts
		if ($target eq 'all')
		{ 
			# Post class
			$class = 'general';
		}

		# System level posts
		elsif ($target eq $realm->{'TxtSystem'})
		{ 
			# Post class
			$class = 'system';

			# Setting of search
			$search{'MPCLASS'} = 'PSTALT';

			# Replace header
			$search{'MNOTICE'} = $tmpl->Pass('TmplPostHeader',
				MSCLASS => 'PSTSYS',
		       		MNOTICE => $realm->{'TagHeadSystem'}
			 );
		}

		# Die roller
		elsif ($target eq 'dice roll')
		{ 
			# Post class
			$class = 'dice';

			# Setting of search
			$search{'MPCLASS'} = 'PSTALT';

			# Replace header
			$search{'MNOTICE'} = $tmpl->Pass('TmplPostHeader',
				MSCLASS => 'PSTDIE',
		       		MNOTICE => $realm->{'TagHeadDice'}
			 );
			
			# Destination
			$dest = (defined($param->{'DPVTE'})) 
			 ? $param->{'USER'}
			 : 'all';
		}
	
		# Private Messages
		elsif ($target ne 'system')
		{
			# Determine class
			# Pull privs
			($sup, $adm, $mod) = $database->DataGetPrivs($param->{'USER'});

			# Assign special permissions to hosts
			$class = (($sup eq 'yes') 
				 || ($adm eq 'yes') 
				 || ($mod eq 'yes')
				 || (defined($param->{'PRINTOUT'}))) ? 'special' : 'private';

			# Setting of search
			$search{'MPCLASS'} = 'PSTALT';

			# Replace header
			$search{'MNOTICE'} = $tmpl->Pass('TmplPostHeader',
				MSCLASS => 'PSTPVT',
		       		MNOTICE => $realm->{'TagHeadPrivate'}
			 );

			# Set target specifically
			$dest = $param->{'TARGET'};
		}
	}


	# Basic post massaging
	# Sanity checks
	$param->{'POST'} = $filter->MakeSane($param->{'POST'}, $realm->{'SetTruncate'});

	# Language filtering
	if ($rlm->{'RealmLanguage'} eq 'restricted')
	{
		# Retreive words
		$database->GetListWords(\@lang, $param->{'ROOM'});

		# Run through
		my $word;

		while ($word = shift(@lang))
		{
			# Run a global replace
			$param->{'POST'} =~ s/$word/$rlm->{'RealmFilter'}/gsi;
			$param->{'CHAR'} =~ s/$word/$rlm->{'RealmFilter'}/gsi;
		}
	}


	
	# Biographical link
	# Inline parameters
	$inline = $param->Crypt(
		BCHAR => $param->{'CHAR'},
		BPOOL => $pup->{'PuppetPool'}
	 );

	# Target
	$window = uc($param->{'CHAR'});
	$window =~ s/\W//gs;


	# Set Search for link
	$search{'MLINK'} = $tmpl->Pass('TagPostLink',
		MLINK   => $realm->{'LnkBio'} . $inline,
		MTARGET => $window
	 );


	# Title
	# Format title
	if (defined($pup->{'PuppetTitle'}))
	{ 
		# Define naming
		$name = ($gummy ne 'NIL')
	         ? $param->{'CHAR'} 
		 : "MLINK$param->{CHAR}</A>";


		# Verify if NAME attribute exists
		if ($pup->{'PuppetTitle'} =~ /NAME/)
		{
			# Prevent tag embed
			$pup->{'PuppetTitle'} =~ s/<NAME>/NAME/g;

			# Replace with puppet name
			$pup->{'PuppetTitle'} =~ s/NAME/$name/g;
			$title = $pup->{'PuppetTitle'};
		}

		# Otherwise prepend name
		else
		{ 
			# Affix the name
			$title = $name . ' ' . $pup->{'PuppetTitle'}; 
		}
	}
	
	# Safety
	$title =  (defined($title)) ? $title : $param->{'CHAR'};

	# Modify once
	$title =~ s/\n/<BR>/gs;


	# Taglines
	# Tagline handling
	$tag =  (defined($pup->{'PuppetTag'}))
        ? $tmpl->Pass('TagPostTagline', MNAME=>$pup->{'PuppetTag'})
	: '';

	# Modify once	
	$tag =~ s/\n/<BR>/gs;
	

	# Gummy Image
	# Set Search
	$gummy = $self->GummyHandler($database, $param, $realm, $pup);

	# Create search fields
	$f_search{'MGUMMY'} = $gummy;
	$i_search{'MGUMMY'} = $gummy;
	$s_search{'MGUMMY'} = (length($gummy) > 1)
	  ? $gummy
	  : "MLINK<IMG SRC=\"$realm->{SetGummyDefault}\" BORDER=\"0\" ALIGN=\"LEFT\" ALT=\"$param->{CHAR}\"></A>";


	# Image formatting
	# Intialization
	$img_lnk = '';
	
	# Need for control tests
	if (defined($pup->{'PuppetPic'}))
	{
		# Define images
		$img_height = $pup->{'PuppetPicHeight'};
		$img_width  = $pup->{'PuppetPicWidth'};
		$img_size   = $pup->{'PuppetPicSize'};


		# Ignore non restricted realms
		# Demension limitations
		if (($rlm->{'RealmImageHeight'} != 0)
		 && ($rlm->{'RealmImageWidth'} != 0))
		{
			# Height handling
			if ($img_height > $rlm->{'RealmImageHeight'})
			{
				# Get average
				$img_avg = $rlm->{'RealmImageHeight'} / $img_height;

				# Forceful resize
				$img_height = int($img_height * $img_avg);
				$img_width  = int($img_width  * $img_avg);
			}

			# Width handling
			if ($img_width > $rlm->{'RealmImageWidth'})
			{
				# Get average
				$img_avg = $rlm->{'RealmImageWidth'} / $img_width;

				# Forceful resize
				$img_height = int($img_height * $img_avg);
				$img_width  = int($img_width  * $img_avg);
			}
		}

		# Image size
		# If everything works
		$img_dis = "<IMG SRC=\"$pup->{'PuppetPic'}\" HEIGHT=\"$img_height\" WIDTH=\"$img_width\" ALIGN=\"left\" BORDER=\"0\">"
		 if (($rlm->{'RealmImageSize'} == 0) || ($img_size < $rlm->{'RealmImageSize'}));
		
		# Text based link only
		$img_lnk = "<A HREF=\"$pup->{PuppetPic}\" TARGET=\"$window\">$realm->{TagLnkImage}<\/A><BR>";
	}

	# Safety
	$img_dis = (defined($img_dis)) ? $img_dis : $img_lnk;

	# Create search fields
	$f_search{'MIMAGE'} = $img_dis;
	$i_search{'MIMAGE'} = $img_lnk;
	$s_search{'MIMAGE'} = $img_lnk;


	# Pre-post location
	# Location specified
	if (defined($param->{'LOCL'}))
	{
		# Check for existence
		# Embed to make nice
		if (($title =~ /LOCALE/)
		 || ($tag =~ /LOCALE/))
		{
			# Replace macro
			$title =~ s/LOCALE/$param->{'LOCL'}/gs;
			$tag   =~ s/LOCALE/$param->{'LOCL'}/gs;
		}

		# Classic behaviour
		else
		{
			# Add additional location tag
			$param->{'POST'} = $tmpl->Pass('TagPostLocale',
				MLOCL => $param->{'LOCL'},
			 	MPOST => $param->{'POST'}
			 );
		}

		# Set class
		$class = 'play';
	}

	# Safety catch
	else
	{
		# No change in class
		if (($title =~ /LOCALE/)
		 || ($tag =~ /LOCALE/))
		{
			# Replace macro
			$title =~ s/LOCALE/$realm->{'TxtLocale'}/gs;
			$tag   =~ s/LOCALE/$realm->{'TxtLocale'}/gs;
		}
	}


	# Easy going modes
	# Correct for normal use
	($tag, $title) = $self->Time(
		$param,
		$realm,
		$rlm->{'RealmDateLoc'},
		$target,
		$tag,
		$tmpl->Pass('TagPostTitle', MNAME=>$title)
	 );

	
	# Complete assignments for normal
	# Full view mode
	$f_search{'MTITLE'} = $title;
	$f_search{'MTAGLN'} = $tag;

	# Imageless view mode
	$i_search{'MTITLE'} = $title;
	$i_search{'MTAGLN'} = $tag;


	# Re-assing for nazi behaviour
	# Reset
	$tag = '';
	
	# Correct for normal use
	($tag, $title) = $self->Time(
		$param,
		$realm,
		$rlm->{'RealmDateLoc'},
		$target,
		$tag,
		$tmpl->Pass('TagPostTitle', MNAME=>$param->{'CHAR'})
	 );
	
	# Spartan mode
	$s_search{'MTITLE'} = $title;
	$s_search{'MTAGLN'} = $tag;


	# Post massage
	# Filtering of the text
	$param->{'POST'} = ($rlm->{'RealmTags'} eq 'restricted')
	 ? $filter->StripHTML($param->{'POST'})
	 : $filter->SafeHTML($param->{'POST'});

	# Large post handling
	# Divide and conquer
	if ((length($param->{'POST'})) > 10240)
	{
		# Cannot allow stray HTML tags
		$param->{'POST'} =~ s/<[^>]*>//gs;

		# Prepend beginning of posts
		$post = substr($param->{'POST'}, 0, 10240);
		push(@posts, "$post ...");

		$position++;

		# Truncate into edible bites
		while (length(substr($param->{'POST'}, ($position * 10240), 10241)) > 10240)
		{
			$post = substr($param->{'POST'}, ($position * 10240), 10240);
			push(@posts, "... $post ...");

			$position++;
		}

		# Append whatever is left
		$post = substr($param->{'POST'}, ($position * 10240), 10240);
		push(@posts, "... $post");
	}
	# Pass as first element of array
	else
	{ 
		# Push onto array
		push(@posts, $param->{'POST'});
	}


	# Clear values
	$param->Cleanup('POST', 'STRMKILL');


	# Determine font colour, face and size
	my ($fontc) = $database->DataGetFontCode($param->{'COLOR'}, 'color');
	my ($fontf) = $database->DataGetFontCode($param->{'FACE'}, 'face');
	my ($fonts) = $database->DataGetFontCode($param->{'SIZE'}, 'size');


	# Cycle though individual sections
	foreach $post (@posts)
	{
		# Handling of links
		$post      =~ s/$/\n/;
		$post      =~ s/(http:\/\/.*?)(\s|$)/<A HREF="$1" TARGET="_blank">$realm->{TagHyperlink}<\/A> /gm;
		$post      =~ s/([\w\.\-]*?\@[\w\.\-]*?\..*?)(\s|$)/<A HREF="mailto:$1">$realm->{TagEmail}<\/A> /gm;

		# Add colors
		$post = $tmpl->Pass('TagPostFont',
			MSIZE  => $fonts,
			MCOLOR => $fontc,
			MFACE  => $fontf,
			MPOST  => $post
		 );

		# Save copy for spartan
		$spartan = $post;


		# Normal posts handler
		# Gummy injection
		$post = $filter->Inject($database, $post);

		# Convert \n to a useable character	
		$post =~ s/\n/<BR>/gs;

		# Create first two elements
		$full      = $tmpl->Pass('TmplPost', %f_search, %search, MPOST=>$post);
		$imageless = $tmpl->Pass('TmplPost', %i_search, %search, MPOST=>$post);


		# Squeaky clean
		# Create spartan format
		$spartan =~ s/<[^>]*>//gs;
		$spartan =~ s/\n/<BR>/gs;

		# Create spartan template
		$spartan  = $tmpl->Pass('TmplPost', %s_search, %search, MPOST=>$spartan);
		

		# Posting
		# Offline posts
		if (defined($param->{'OFFLINE'}))
		{
			# Insert into posts
			$database->Quick("INSERT INTO Post
				(RealmName,
				 PuppetName,
				 PuppeteerLogin,
				 PostTo,
				 PostClass,
				 PostFullFormat,
				 PostImagelessFormat,
				 PostPartialFormat,
				 PostCmd)		 
				VALUES(?,?,?,?,?,?,?,?,?)",
	
			 $param->{'ROOM'}, 
			 $param->{'CHAR'}, 
			 $param->{'USER'}, 
			 $dest, 
			 $class,
			 $full, 
			 $imageless, 
			 $spartan,
			 $cmd);


		 	# Replace
			$full      =~ s/PSTOLN/PSTPRI/s;
			$imageless =~ s/PSTOLN/PSTPRI/s;
			$spartan   =~ s/PSTOLN/PSTPRI/s;
			

			# Insert into posts
			$database->Quick("INSERT INTO Offline
				(RealmName,
				 PuppetName,
				 PuppeteerLogin,
				 OfflineFullFormat,
				 OfflineImagelessFormat,
				 OfflinePartialFormat)
				VALUES(?,?,?,?,?,?)",
	
			 $param->{'ROOM'}, 
			 $param->{'CHAR'}, 
			 $param->{'USER'}, 
			 $full, 
			 $imageless, 
			 $spartan);


		 	# Email notification
			if ($count < 1)
			{
				# Specific data handlers
				my $send;				# Sending handle
				my $w_handle;				# Watch statement
				my $w_res;				# Watch results

				my %system;				# System hash
			

				# Prepare and execute
				$database->Pull(\$w_handle, "SELECT
					  p.PuppeteerName  AS \"Name\",
					  p.PuppeteerEmail AS \"Email\"
					FROM Puppeteer p, Watch w
					WHERE w.PuppeteerLogin = p.PuppeteerLogin
					  AND w.PuppetName=?
					  AND w.PuppetPool=?
					  AND w.RealmName=?
					ORDER BY p.PuppeteerEmail",

				 $pup->{'PuppetName'},
				 $pup->{'PuppetPool'},
				 $param->{'ROOM'}
				);


				# Necessary
				if ($w_res = $w_handle->fetchrow_hashref())
				{
					# Link with system hash
					$database->GetHashSystem(\%system);
			
					# Pull
					($notice) = $database->DocumentGetMailOffline();


					# Send Mail
					# Handle initiation
					$send = tie(*MAIL, 'Ethereal::Mail', $database);

					# Initial setup
					$send->SetSubject("$system{SetBoardPrefix} $param->{CHAR} - $param->{'ROOM'}");
					$send->SetSearch(
						MFROM => $param->{'CHAR'},
						MMAIL => "$system{SetInfoContactName}\n$system{SetInfoContactAddress}",
						MROOM => $param->{'ROOM'},
						MPOST => $offline 
					 );

					# Default Recipients
					$send->AddTo("$system{SetSenderName} \<$system{SetSenderAddress}\>");

					# True Recipients
					do
					{
						# BCC to hide identity
						$send->AddBCC("$w_res->{Name} \<$w_res->{Email}\>");

					} while ($w_res = $w_handle->fetchrow_hashref());
	
					# Add message
					print MAIL $notice;

					# Close and send
					close(MAIL);
				}

				# Finish up
				$w_handle->finish();
			}
		}

		# Active realms
		else
		{
			# Normal messages
			if (($target eq 'all')
			 || ($target eq 'dice roll'))
			{
				# Insert into database
				$database->Quick("INSERT INTO Post
					(RealmName,
					 PuppetName,
					 PuppeteerLogin,
					 PostTo,
					 PostClass,
					 PostFullFormat,
					 PostImagelessFormat,
					 PostPartialFormat,
					 PostCmd)		 
					VALUES(?,?,?,?,?,?,?,?,?)",
	
				 $param->{'ROOM'}, 
				 $param->{'CHAR'}, 
				 $param->{'USER'}, 
				 $dest, 
				 $class,
				 $full, 
				 $imageless, 
				 $spartan,
				 $cmd);
			}

			# System wide
			elsif ($target eq $realm->{'TxtSystem'})
			{
				# Prepare and execute
				# Design
				$database->Pull(\$t_handle, "SELECT RealmName
					FROM Post 
					WHERE PostTimestamp > getInterval('10 Minutes')
					GROUP BY RealmName");


				# Run through each realm
				while (($res) = $t_handle->fetchrow())
				{
					# Insert into database
					$database->Quick("INSERT INTO Post 
						(RealmName,
						 PuppetName,
						 PuppeteerLogin,
						 PostTo,
						 PostClass,
						 PostFullFormat,
						 PostImagelessFormat,
						 PostPartialFormat,
						 PostCmd)
						VALUES(?,?,?,?,?,?,?,?,?)", 
					 $res, 
					 $param->{'CHAR'}, 
					 $param->{'USER'}, 
					 $dest,
					 $class,
					 $full, 
					 $imageless, 
					 $spartan,
					 $cmd);
				}

				# End statement
				$t_handle->finish();
			}

			# Insert message to be returned to poster
			elsif ($param->{'TARGET'} ne 'system')
			{
				# Send to target
				$database->Quick("INSERT INTO Post
					(RealmName,
					 PuppetName,
					 PuppeteerLogin,
					 PostTo,
					 PostClass,
					 PostFullFormat,
					 PostImagelessFormat,
					 PostPartialFormat,
					 PostCmd)
					VALUES(?,?,?,?,?,?,?,?,?)", 

				 $param->{'ROOM'}, 
				 $param->{'CHAR'}, 
				 $param->{'USER'}, 
				 $dest,
				 $class,
				 $full, 
				 $imageless, 
				 $spartan,
				 $cmd);


				# No doubling for self posts
				if ($dest ne $param->{'SOURCE'})
				{
					# Reset target
					$dest  = $param->{'SOURCE'};
					$class = 'duplicate';

					# Send back owner
					$database->Quick("INSERT INTO Post
						(RealmName,
						 PuppetName,
						 PuppeteerLogin,
						 PostTo,
						 PostClass,
						 PostFullFormat,
						 PostImagelessFormat,
						 PostPartialFormat,
						 PostCmd)
						VALUES(?,?,?,?,?,?,?,?,?)",

					 $param->{'ROOM'}, 
					 $param->{'CHAR'}, 
					 $param->{'USER'}, 
					 $dest, 
					 $class,
					 $full, 
					 $imageless, 
					 $spartan,
					 $cmd);
				}
			}
		}

		# Increment
		$count++;
	}


	# End of handles
	$r_handle->finish();
	$p_handle->finish();
}


#####################
# Date/Time stamp
#
# Functionality that allows the host to decide more or less where the date and
# time stamp to be placed.   Depending on the locale, different output will
# be generated.

sub Time
{
	#####################
	# Data members
	my $self     = shift;					# Self referencing link
	my $param    = shift;					# Parameter handle
	my $realm    = shift;					# Realm hash
	my $pref     = shift;					# Preffered position
	my $trgt     = shift;					# Target audience
	my $tag      = shift;					# String to handle
	my $title    = shift;					# String to handle

	my $database = $self->{'HANDLE'};			# Database handle
	my $tmpl     = $self->{'TEMPLATE'};			# Template

	my $pesky;						# Handling of global posts
	my $time;						# Preffered position

	my $from;						# Who it is from
	my $room;						# Room of origin
	my $sep;						# Seperator
	my $to;							# Going to
	my $tmp;						# Temporary holder


	#####################
	# Program area	

	# Determine audience
	# Simply the date for these
	unless (($trgt ne 'all') 
	    && ($trgt ne 'system')
	    && ($trgt ne 'dice roll')
	    && ($trgt ne $realm->{'TxtSystem'}))
	{
		# No need to change values
		if (($tag !~ /DATETIME/)
		 && ($title !~ /DATETIME/))
		{
			# Change format
			if ($pref eq 'bottom')
			{
				# Classical behaviour
				$tag .= ' ' . $tmpl->Pass('TagPostTagline', MNAME=>'DATETIME');
			}
			elsif ($pref eq 'top')
			{
				# Alternate behaviour
				$title = $tmpl->Pass('TagPostTopline', MNAME=>'DATETIME') .'<BR>'. $title;
			}
		}
	}

	# Date with identification
	else
	{
		# Initialize
		$pesky = '';

		# Adjust for tags
		# Insert additional spaces
		$room = $realm->{'TagPuppetPMRealm'} . ' ';
		$from  = $realm->{'TagPuppetPMFrom'} . ' ';
		$to    = $realm->{'TagPuppetPMTo'}   . ' ';
		$sep   = ' ' . $realm->{'TagPuppetPMSep'}  . ' ';
		
		# Change format
		# Generate regular string
		if ($pref eq 'bottom')
		{
			# Pesky handling
			$pesky = $sep . $room . $param->{'ROOM'}
		          if (defined($param->{'PESKY'}));

			# Set string
			$tag .= $tmpl->Pass('TagPostTagline',
			     MNAME=> $from . $param->{'CHAR'}    . $sep
			          .  $to   . $param->{'OTARGET'}
			          .  $pesky); 
				     
			# Append timestamp if required
			$tag .= $tmpl->Pass('TagPostTagline', MNAME=>'DATETIME')
			 if (($tag !~ /DATETIME/) && ($title !~ /DATETIME/));
		}

		# Generate stripped string
		else
		{
			# Pesky handling
			$pesky = $sep . $room . $param->{'ROOM'}
			  if (defined($param->{'PESKY'}));

			# Set string
			$tmp =    $tmpl->Pass('TagPostTopline',
				MNAME=> $from . $param->{'CHAR'}    . $sep
			             .  $to   . $param->{'OTARGET'}
			             .  $pesky); 

			# Append timestamp if required
			$tmp .= $tmpl->Pass('TagPostTime', MNAME=>$sep . 'DATETIME')
			 if (($tag !~ /DATETIME/) && ($title !~ /DATETIME/));

			# Define title
			$title = $tmp . '<BR>' . $title;
		}
	}


	# Time handling
	# Pull time
	($time) = $database->GetDateFormat($param->{'ROOM'});

	# Run through template
	$time = $tmpl->Pass('TagPostTime', MNAME=>$time);

	
	# Final replacement
	$tag  =~ s/DATETIME/$time/gs;
	$title =~ s/DATETIME/$time/gs;

	# Return variables
	# With additional HTML for ignore
	return $tag, $title;
}

1;								# Return true
